import { Expose, plainToInstance } from 'class-transformer';
import { E as ENTITY_KEY } from '../../dist/entity-6fbc681a.esm.js';
export { U as Union } from '../../dist/entity-6fbc681a.esm.js';
import { defaultMetadataStorage } from 'class-transformer/cjs/storage.js';
import get from 'lodash/get';
import set$1 from 'lodash/set';

function _initializerDefineProperty(target, property, descriptor, context) {
  if (!descriptor) return;
  Object.defineProperty(target, property, {
    enumerable: descriptor.enumerable,
    configurable: descriptor.configurable,
    writable: descriptor.writable,
    value: descriptor.initializer ? descriptor.initializer.call(context) : void 0
  });
}

function _applyDecoratedDescriptor(target, property, decorators, descriptor, context) {
  var desc = {};
  Object.keys(descriptor).forEach(function (key) {
    desc[key] = descriptor[key];
  });
  desc.enumerable = !!desc.enumerable;
  desc.configurable = !!desc.configurable;
  if ('value' in desc || desc.initializer) {
    desc.writable = true;
  }
  desc = decorators.slice().reverse().reduce(function (desc, decorator) {
    return decorator(target, property, desc) || desc;
  }, desc);
  if (context && desc.initializer !== void 0) {
    desc.value = desc.initializer ? desc.initializer.call(context) : void 0;
    desc.initializer = undefined;
  }
  if (desc.initializer === void 0) {
    Object.defineProperty(target, property, desc);
    desc = null;
  }
  return desc;
}

var _dec, _class, _descriptor;
let WithMeta = (_dec = Expose(), (_class = class WithMeta {
  constructor() {
    _initializerDefineProperty(this, "_meta", _descriptor, this);
  }
}, (_descriptor = _applyDecoratedDescriptor(_class.prototype, "_meta", [_dec], {
  configurable: true,
  enumerable: true,
  writable: true,
  initializer: null
})), _class));

class Entity {
  constructor(params) {
    this.type = void 0;
    this.id = void 0;
    this.data = void 0;
    this.path = void 0;
    this.meta = void 0;
    this.id = params.id;
    this.type = params.type;
    this.meta = params.meta;
  }
  get entityTypeId() {
    return this.type.split("--")[0];
  }
  get bundleId() {
    return this.type.split("--")[1];
  }
}

class EntityResolver {
  constructor(httpClient, options) {
    this.httpClient = void 0;
    this.entityMap = new Map();
    this.options = void 0;
    this.options = options;
    this.httpClient = httpClient;
  }
  async execute({
    path,
    data,
    target
  }) {
    const targetEntities = Reflect.getMetadata(ENTITY_KEY, target);
    if (!targetEntities) return this.entityMap;
    const targetExposedProps = defaultMetadataStorage.getExposedMetadatas(target);
    const exposedEntityProps = targetExposedProps.filter(({
      propertyName
    }) => targetEntities.has(propertyName ?? ""));
    const incomingPath = `${path}${path?.length ? "." : ""}`;
    for (const {
      propertyName,
      options
    } of exposedEntityProps) {
      const dataAccessorName = options.name || propertyName;
      if (!dataAccessorName) throw new Error("No data accessor property. Not possible to read entities from resolved data");
      const dataResource = get(data, dataAccessorName);
      if (dataResource === undefined || !Array.isArray(dataResource) && !dataResource.type && !dataResource.id || Array.isArray(dataResource) && !dataResource.length) continue;
      if (propertyName) {
        const {
          options,
          reflectedType,
          typeFunction
        } = defaultMetadataStorage.findTypeMetadata(target, propertyName);
        const type = typeFunction();
        const subTypes = options.discriminator?.subTypes;
        let reflectTypeName = reflectedType?.name;
        const entityDecoratorOptions = targetEntities.get(propertyName ?? "");
        if (!reflectTypeName) {
          reflectTypeName = typeFunction().constructor.name;
        } else if (reflectTypeName !== "Object" && reflectTypeName !== "Array") {
          reflectTypeName = reflectedType.constructor.name;
        }
        if (reflectTypeName === "Array" && Array.isArray(dataResource)) {
          const baseParams = {
            incomingPath,
            dataAccessorName,
            entityDecoratorOptions,
            parentData: data
          };
          if (subTypes) {
            const dataResourceList = dataResource.filter(resourceEntity => subTypes.find(type => type.name === resourceEntity.type)?.value);
            await this.arrayEntityRequest({
              typeOrTypes: subTypes,
              dataResourceList,
              ...baseParams
            });
          } else {
            await this.arrayEntityRequest({
              typeOrTypes: type,
              dataResourceList: dataResource,
              ...baseParams
            });
          }
        } else if ((reflectTypeName === "Object" || reflectTypeName === "Function") && !Array.isArray(dataResource)) {
          const shouldResolve = await this.conditionalResolver({
            entityDecoratorOptions,
            meta: dataResource.meta,
            parentData: data,
            dataAccessorName
          });
          if (!shouldResolve) continue;
          const params = {
            subTypes,
            type,
            incomingPath,
            dataAccessorName
          };
          await this.objectTypeResolver(params, dataResource, entityDecoratorOptions?.parentEntityHolder ?? false);
        }
      }
    }
    return this.entityMap;
  }
  async conditionalResolver({
    entityDecoratorOptions,
    dataAccessorName,
    parentData,
    meta
  }) {
    if (entityDecoratorOptions?.conditionalResolver) {
      const hasIndexMatch = dataAccessorName.match(/\[(.*?)]/);
      const sanitizedMatch = hasIndexMatch?.[0].replace("[", "").replace("]", "");
      const idx = sanitizedMatch ? parseInt(sanitizedMatch) : undefined;
      const shouldResolve = entityDecoratorOptions.conditionalResolver(this.options?.context ?? {}, meta || {}, idx);
      if (!shouldResolve) {
        set$1(parentData, dataAccessorName, idx !== undefined ? {} : undefined);
      }
      return shouldResolve;
    }
    return true;
  }
  async objectTypeResolver({
    incomingPath,
    type,
    subTypes,
    dataAccessorName
  }, dataResource, parentEntityHolder) {
    const entity = this.createEntity(dataResource);
    entity.path = `${incomingPath}${dataAccessorName}`;
    if (parentEntityHolder) {
      entity.data = dataResource;
    } else {
      entity.data = await this.httpClient.get(`/${entity.entityTypeId}/${entity.bundleId}/${entity.id}`);
    }
    this.entityMap.set(entity.path, entity);
    if (!entity.data) throw new Error("Entity has no resolved data");
    if (subTypes) {
      const correspondingSubType = subTypes.find(subType => subType.name === entity.type);
      if (!correspondingSubType) throw new Error("Error retrieving subtype for object entity union");
      await this.execute({
        target: correspondingSubType.value,
        data: entity.data,
        path: entity.path
      });
    } else {
      await this.execute({
        target: type,
        data: entity.data,
        path: entity.path
      });
    }
  }
  async arrayEntityRequest({
    typeOrTypes,
    dataResourceList,
    incomingPath,
    dataAccessorName,
    entityDecoratorOptions,
    parentData
  }) {
    const resolveDataPromises = [];
    for (const [idx, resourceEntity] of dataResourceList.entries()) {
      let type;
      if (Array.isArray(typeOrTypes)) {
        type = typeOrTypes.find(subType => subType.name === resourceEntity.type)?.value;
      } else {
        type = typeOrTypes;
      }
      if (!type) {
        throw new Error("Type for resource entity was not found");
      }
      const dataAccessor = `${dataAccessorName}[${idx}]`;
      const shouldResolve = await this.conditionalResolver({
        entityDecoratorOptions,
        meta: resourceEntity.meta,
        parentData: parentData,
        dataAccessorName: dataAccessor
      });
      if (!shouldResolve) continue;
      const entity = this.createEntity(resourceEntity);
      entity.path = `${incomingPath}${dataAccessor}`;
      this.entityMap.set(entity.path, entity);
      if (entityDecoratorOptions?.parentEntityHolder) {
        entity.data = resourceEntity;
        const data = {
          type,
          entity
        };
        resolveDataPromises.push(Promise.resolve(data));
      } else {
        resolveDataPromises.push(this.httpClient.get(`/${entity.entityTypeId}/${entity.bundleId}/${entity.id}`).then(data => {
          const retrievedEntity = this.entityMap.get(entity.path);
          if (!retrievedEntity) throw new Error("Error retrieving entity");
          retrievedEntity.data = data;
          return {
            type: type,
            entity: retrievedEntity
          };
        }));
      }
    }
    const resolvedData = await Promise.all(resolveDataPromises);
    const concurrentPromises = [];
    for (const item of resolvedData) {
      if (!item) continue;
      const {
        entity,
        type
      } = item;
      if (!entity.data) throw new Error("Entity has no resolved data");
      concurrentPromises.push(this.execute({
        target: type,
        data: entity.data,
        path: entity.path
      }));
    }
    await Promise.all(concurrentPromises);
  }
  createEntity(dataResource) {
    return new Entity({
      type: dataResource.type,
      id: dataResource.id,
      meta: dataResource.meta
    });
  }
}

/* eslint-disable @typescript-eslint/no-var-requires */
const set = require("lodash/set");
class EntityResolverService {
  constructor(httpClient) {
    this.httpClient = void 0;
    this.httpClient = httpClient;
  }
  async get(target, data, options) {
    const entityResolver = new EntityResolver(this.httpClient, options);
    const entityMap = await entityResolver.execute({
      target,
      data,
      options,
      path: ""
    });
    const rebuiltData = this.rebuildResolvedData(entityMap, data) ?? {};
    if (options?.raw) {
      return rebuiltData;
    }
    return plainToInstance(target, rebuiltData, {
      excludeExtraneousValues: true
    });
  }
  rebuildResolvedData(entityMap, data) {
    let rebuiltData = structuredClone(data);
    for (const entity of entityMap.values()) {
      rebuiltData = set(data, entity.path, {
        ...entity.data,
        _meta: entity.meta
      });
    }
    return rebuiltData;
  }
}

export { EntityResolverService, WithMeta };
